import pytest

from uberspace_takeout.storage import LocalMoveStorage
from uberspace_takeout.storage import Storage
from uberspace_takeout.storage import TarStorage


@pytest.mark.parametrize("mode", ["takeout", "takein",])
def test_storage_ctor(mode):
    s = Storage("dest.tar.gz", mode)
    assert s.destination == "dest.tar.gz"
    assert s.mode == mode


def test_storage_ctor_mode():
    with pytest.raises(Exception) as ex:
        Storage("dest.tar.gz", "foo")

    assert "Invalid mode foo" in str(ex)


@pytest.fixture
def test_file(tmp_path):
    path = tmp_path / "some_file.txt"

    with path.open("w") as f:
        f.write(u"some file text")

    return path


@pytest.fixture
def test_file2(tmp_path):
    path = tmp_path / "some_file2.txt"

    with path.open("w") as f:
        f.write(u"some file text2")

    return path


@pytest.fixture
def test_dir(tmp_path):
    (tmp_path / "some_dir/some_subdir").mkdir(parents=True)

    with (tmp_path / "some_dir/file.txt").open("w") as f:
        f.write(u"some file text")

    with (tmp_path / "some_dir/some_subdir/file2.txt").open("w") as f:
        f.write(u"some file text 2")

    return tmp_path / "some_dir"


storages = [
    TarStorage,
    LocalMoveStorage,
]


@pytest.mark.parametrize("storage", storages)
def test_storage_text(storage, tmp_path):
    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        s.store_text("some text 1", "simple_text.txt")
        s.store_text("some text 2", "subdir/bla/simple_text.txt")

    with storage(tmp_path / "test.tar.gz", "takein") as s:
        assert s.unstore_text("simple_text.txt") == "some text 1"
        assert s.unstore_text("subdir/bla/simple_text.txt") == "some text 2"


@pytest.mark.parametrize("storage", storages)
def test_storage_file(storage, tmp_path, test_file, test_file2):
    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        s.store_file(test_file, "simple_file.txt")
        s.store_file(test_file2, "subdir/bla/simple_file.txt")

    with storage(tmp_path / "test.tar.gz", "takein") as s:
        assert s.unstore_text("simple_file.txt") == "some file text"
        # read twice to check hat we don't break the tarinfo objects
        assert s.unstore_text("subdir/bla/simple_file.txt") == "some file text2"
        assert s.unstore_text("subdir/bla/simple_file.txt") == "some file text2"

        s.unstore_file("simple_file.txt", tmp_path / "extracted.txt")

        with (tmp_path / "extracted.txt").open() as f:
            assert f.read() == "some file text"


@pytest.mark.parametrize("storage", storages)
def test_storage_file_eexists(storage, tmp_path, test_file):
    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        s.store_file(test_file, "simple_file.txt")

        test_file.touch()

        with pytest.raises(FileExistsError):
            s.store_file(test_file, "simple_file.txt")


@pytest.mark.parametrize("storage", storages)
def test_storage_file_noent(storage, tmp_path, test_file):
    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        pass

    with storage(tmp_path / "test.tar.gz", "takein") as s:
        with pytest.raises(FileNotFoundError):
            s.unstore_file("simple_file.txt", tmp_path / "somefile")


@pytest.mark.parametrize("storage", storages)
def test_storage_directory(storage, tmp_path, test_dir):
    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        s.store_directory(test_dir, "dir")

    with storage(tmp_path / "test.tar.gz", "takein") as s:
        assert s.unstore_text("dir/some_subdir/file2.txt") == "some file text 2"

        s.unstore_directory("dir", tmp_path / "new_dir")

        with (tmp_path / "new_dir" / "file.txt").open() as f:
            assert f.read() == "some file text"

        with (tmp_path / "new_dir" / "some_subdir/file2.txt").open() as f:
            assert f.read() == "some file text 2"


@pytest.mark.parametrize("storage", storages)
def test_storage_directory_eexists(storage, tmp_path, test_dir):
    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        s.store_directory(test_dir, "dir")

        test_dir.mkdir(exist_ok=True)

        with pytest.raises(FileExistsError):
            s.store_directory(test_dir, "dir")


@pytest.mark.parametrize("storage", storages)
def test_storage_directory_noent(storage, tmp_path):
    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        pass

    with storage(tmp_path / "test.tar.gz", "takein") as s:
        with pytest.raises(FileNotFoundError):
            s.unstore_directory("dir", tmp_path / "new_dir")


@pytest.mark.parametrize("storage", storages)
def test_storage_list_files(storage, tmp_path, test_dir):
    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        s.store_directory(test_dir, "dir")
        assert sorted(s.list_files("dir")) == ["file.txt", "some_subdir"]
        assert sorted(s.list_files("dir/some_subdir")) == ["file2.txt"]


@pytest.mark.parametrize("storage", storages)
def test_storage_list_files_noent(storage, tmp_path, test_dir):
    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        pass

    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        with pytest.raises(FileNotFoundError):
            s.list_files("dir")


@pytest.mark.parametrize("storage", storages)
def test_storage_slashes(storage, tmp_path, test_file):
    with storage(tmp_path / "test.tar.gz", "takeout") as s:
        s.store_text("some text 3", "/subdir/bla/simple_text2.txt")
        s.store_file(test_file, "/subdir/bla/simple_file2.txt")

    with storage(tmp_path / "test.tar.gz", "takein") as s:
        assert s.unstore_text("/subdir/bla/simple_text2.txt") == "some text 3"
        assert s.unstore_text("/subdir/bla/simple_file2.txt") == "some file text"
        assert s.unstore_text("subdir/bla/simple_text2.txt") == "some text 3"
        assert s.unstore_text("subdir/bla/simple_file2.txt") == "some file text"


def test_localmovestorage_file_removed(tmp_path, test_file):
    with LocalMoveStorage(tmp_path / "test.tar.gz", "takeout") as s:
        assert test_file.exists()
        s.store_file(test_file, "simple_file.txt")
        assert not test_file.exists()


def test_localmovestorage_directory_removed(tmp_path, test_dir):
    with LocalMoveStorage(tmp_path / "test.tar.gz", "takeout") as s:
        assert test_dir.exists()
        s.store_directory(test_dir, "dir")
        assert not test_dir.exists()



import re
import os
import subprocess


class TakeoutItem:
    description = None

    def __init__(self, username, hostname, storage):
        self.username = username
        self.hostname = hostname
        self.storage = storage

    def takeout(self):
        raise NotImplementedError()

    def takein(self):
        raise NotImplementedError()

    def is_active(self):
        return True

    def run_command(self, cmd, input_text=None):
        env = os.environ.copy()
        env["PATH"] = "/usr/local/bin/:" + env["PATH"]
        env.pop("SUDO_USER", None)

        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdin=subprocess.PIPE,
            env=env,
            universal_newlines=True,  # get a string, not bytes
        )
        out, _ = p.communicate(input_text)

        return [l for l in out.split("\n") if l]

    def run_uberspace(self, *cmd):
        return self.run_command(["uberspace"] + list(cmd))


class PathItem(TakeoutItem):
    storage_path = None

    def takeout(self):
        self.storage.store_directory(self.path(), self.storage_path)

    def takein(self):
        self.storage.unstore_directory(self.storage_path, self.path())


class UberspaceVersionMixin:
    uberspace_version = None

    @property
    def current_uberspace_version(self):
        with open("/etc/centos-release") as f:
            text = f.read()
            # looks like "CentOS release 6.10 (Final)"
            centos_release = re.search(r"release ([0-9])+\.", text).groups()[0]
            return int(centos_release)

    def is_active(self):
        return self.current_uberspace_version == int(self.uberspace_version)



import re

from .base import TakeoutItem
from .base import UberspaceVersionMixin


class U7Mixin(UberspaceVersionMixin):
    uberspace_version = 7


def convert_legacy_domain(domain):
    """
    convert legacy domains to new .uber.space ones, e.g.

         luto.cygnus.uberspace.de   =>      luto.uber.space
      ep.luto.cygnus.uberspace.de   =>   ep.luto.uber.space
    """

    if domain.endswith('.uberspace.de'):
        # strip suffix off a full domain, e.g.
        #   luto.cygnus.uberspace.de  =>  luto
        domain = re.sub(r'\.[a-z]+\.uberspace\.de$', '', domain) + '.uber.space'

    return domain

class DomainItem(U7Mixin, TakeoutItem):
    area = None

    def takeout(self):
        domains = self.run_uberspace(self.area, "domain", "list")
        domains = set(domains) - {
            self.username + ".uber.space",
            self.username + "." + self.hostname,
        }
        self.storage.store_text("\n".join(domains), self.storage_path)

    def takein(self):
        text = self.storage.unstore_text(self.storage_path)
        for domain in (d for d in text.split("\n") if d):
            if domain.startswith("*."):
                print("cannot add wildcard domain on: " + domain)
                continue
            if domain.endswith('.uberspace.de'):
                print("user.host.uberspace.de domains are not supported, rewriting to .uber.space")
                domain = convert_legacy_domain(domain)
            self.run_uberspace(self.area, "domain", "add", domain)


class WebDomains(DomainItem):
    description = "Web Domains"
    area = "web"
    storage_path = "conf/domains-web"


class MailDomains(DomainItem):
    description = "Mail Domains"
    area = "mail"
    storage_path = "conf/domains-mail"


class FlagItem(U7Mixin, TakeoutItem):
    """
    A flag like "is the spamfilter enabled?". It provide a status/enable/disable
    interface via a uberspace sub-command. Provide the uberspace command without
    leading "uberspace" in `cmd` (e.g. ['web', 'log', 'access']).
    """

    def takeout(self):
        out = self.run_uberspace(*(self.cmd + ["status"]))
        if "enabled" in out:
            status = "enable"
        else:
            status = "disable"

        self.storage.store_text(status, self.storage_path)

    def takein(self):
        try:
            data = self.storage.unstore_text(self.storage_path)
        except FileNotFoundError:
            return

        if data not in ("enable", "disable"):
            raise Exception(
                'invalid "uberspace {}" value: {}, expected "enabled" or "disabled".'.format(
                    " ".join(self.cmd), data,
                )
            )

        self.run_uberspace(*(self.cmd + [data]))


class AccessLogItem(FlagItem):
    description = "Setting: Access-Log"
    cmd = ["web", "log", "access"]
    storage_path = "conf/log-access"


class ApacheErrorLogItem(FlagItem):
    description = "Setting: Apache-Error-Log"
    cmd = ["web", "log", "apache_error"]
    storage_path = "conf/log-apache_error"


class PhpErrorLogItem(FlagItem):
    description = "Setting: PHP-Error-Log"
    cmd = ["web", "log", "php_error"]
    storage_path = "conf/log-php_error"


class SpamfilterLogItem(FlagItem):
    description = "Setting: Spamfilter"
    cmd = ["mail", "spamfilter"]
    storage_path = "conf/spamfilter-enabled"


class ToolVersions(U7Mixin, TakeoutItem):
    description = "Setting: Tool Versions"

    def takeout(self):
        tools = self.run_uberspace("tools", "version", "list")

        for tool in tools:
            tool = tool.lstrip("- ")
            out = self.run_uberspace("tools", "version", "show", tool)
            version = re.search(r"'([0-9\.]+)'", out[0]).groups()[0]
            self.storage.store_text(version, "conf/tool-version/" + tool)

    def takein(self):
        try:
            tools = self.storage.list_files("conf/tool-versions")
        except FileNotFoundError:
            return

        for tool in tools:
            version = self.storage.unstore_text("conf/tool-versions/" + tool)
            self.run_uberspace("tools", "version", "use", tool, version)



import configparser

from .base import PathItem
from .base import TakeoutItem
from uberspace_takeout.exc import TakeoutError


class TakeoutMarker(TakeoutItem):
    description = "Takeout Marker (internal)"

    def takeout(self):
        self.storage.store_text("uberspace_takeout", ".uberspace_takeout")

    def takein(self):
        content = self.storage.unstore_text(".uberspace_takeout")

        if content != "uberspace_takeout":
            raise Exception("input is not a takeout.")


class Homedir(PathItem):
    description = "Homedirectory"
    storage_path = "home/"

    def path(self):
        return "/home/" + self.username


class Www(PathItem):
    description = "Documentroot"
    storage_path = "www/"

    def path(self):
        return "/var/www/virtual/" + self.username


class Cronjobs(TakeoutItem):
    description = "Cronjobs"

    def takeout(self):
        cronjobs = self.run_command(["crontab", "-l"])

        if len(cronjobs) == 1 and cronjobs[0].startswith("no crontab for"):
            text = ""
        else:
            text = "\n".join(cronjobs) + "\n"

        self.storage.store_text(text, "conf/cronjobs")

    def takein(self):
        text = self.storage.unstore_text("conf/cronjobs")
        self.run_command(["crontab", "-"], input_text=text)


class MySQLPassword(TakeoutItem):
    description = "MySQL password"

    @property
    def _my_cnf_path(self):
        return "/home/" + self.username + "/.my.cnf"

    def _open_my_cnf(self, section):
        config = configparser.ConfigParser(interpolation=None)
        config.read(self._my_cnf_path, encoding="utf-8")
        return config

    def _read_my_cnf_password(self, section):
        raw_pw = self._open_my_cnf(section)[section]["password"]
        return raw_pw.partition("#")[0].strip().strip('"')

    def _check_password(self, password):
        if len(password) < 12:
            msg = (
                "Your current MySQL password does not satisfy our policy requirements:"
                " it is shorter than 12 characters."
                " Please set a sufficiently long one - as described under"
                " https://wiki.uberspace.de/database:mysql#passwort_aendern"
                " - or run this script with '--skip-item MySQLPassword'."
            )
            raise TakeoutError(msg)

    def _write_my_cnf_password(self, section, password):
        config = self._open_my_cnf(section)
        config[section]["password"] = password

        with open(self._my_cnf_path, "w") as f:
            config.write(f)

    def _set_password(self, suffix):
        password = self.storage.unstore_text("conf/mysql-password-client" + suffix)
        self.run_command(
            [
                "mysql",
                "--defaults-group-suffix=" + suffix,
                "-e",
                "SET PASSWORD = PASSWORD('" + password + "')",
            ]
        )
        self._write_my_cnf_password("client" + suffix, password)

    def takeout(self):
        password = self._read_my_cnf_password("client")
        self.storage.store_text(password, "conf/mysql-password-client")
        self._check_password(password)

    def takein(self):
        self._set_password("")



import argparse
import datetime
import getpass
import sys

from . import __version__ as version
from . import Takeout


def main():
    username = getpass.getuser()
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H_%M_%S")
    default_tar_path = "takeout_{}_{}.tar.bz2".format(username, timestamp)

    p = argparse.ArgumentParser()
    p.add_argument("action", choices=["takeout", "takein", "items"])
    p.add_argument("--username", default=username)
    p.add_argument("--skip-item", action="append", default=[])
    p.add_argument("--tar-file", default=default_tar_path)
    p.add_argument("--version", action="version", version=version)
    args = p.parse_args()

    tar_path = args.tar_file
    t = Takeout()

    if args.action == "takeout":
        if tar_path == "-":
            tar_path = "/dev/stdout"
            sys.stdout = sys.stderr

        print("writing " + tar_path)
        t.takeout(tar_path, args.username, args.skip_item)

    elif args.action == "takein":
        if tar_path == "-":
            tar_path = "/dev/stdin"

        print("reading " + tar_path)
        t.takein(tar_path, args.username, args.skip_item)

    elif args.action == "items":
        print(
            "\n".join(
                i.__name__.ljust(25, " ") + i.description for i in Takeout.takeout_menu
            )
        )
        return 0

    else:
        raise NotImplementedError()

    if t.errors:
        print()
        for item_class, errors in t.errors.items():
            print(f"[ERROR] {item_class}:")
            for error in errors:
                print(f"- {error}")
        print()
        return 1

    else:
        return 0


if __name__ == "__main__":
    sys.exit(main())



import subprocess

from uberspace_takeout import __version__


def test_cli_version(capfdbinary):
    command = ["uberspace-takeout", "--version"]
    res = subprocess.run(command)
    assert res.returncode == 0
    out, err = capfdbinary.readouterr()
    version = out.decode("utf-8").strip()
    assert version == __version__
